@@ -1,33 +1,51 @@
-/*---------------------------------------------------------------------------
-  Module FmPopup
+/*
+ *  This file is part of X-File Manager XFM
+ *  ----------------------------------------------------------------------
+  FmPopup.c
 
   (c) Simon Marlow 1990-92
   (c) Albert Graef 1994
 
-  Routines for creating and managing popup forms and dialog boxes
-
   modified 1-29-95 by rodgers@lvs-emh.lvs.loral.com (Kevin M. Rodgers)
   to add filtering of icon/text directory displays by a filename filter.
 
   modified 7-1997 by strauman@sun6hft.ee.tu-berlin.de to add
-  different enhancements (see README-NEW).
+  different enhancements (see README-1.4).
 
----------------------------------------------------------------------------*/
+  modified 2004,2005,2006 by Bernhard R. Link (see Changelog)
+
+  Routines for creating and managing popup forms and dialog boxes
+ *  ----------------------------------------------------------------------
+ *  This program is free software; you can redistribute it and/or modify
+ *  it under the terms of the GNU General Public License as published by
+ *  the Free Software Foundation; either version 2 of the License, or
+ *  (at your option) any later version.
+
+ *  This program is distributed in the hope that it will be useful,
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *  GNU General Public License for more details.
+
+ *  You should have received a copy of the GNU General Public License
+ *  along with this program; if not, write to the Free Software
+ *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
+ */
+#include <xfmconfig.h>
 
 #include <stdio.h>
 #include <string.h>
 #include <unistd.h>
+#include <assert.h>
 
 #include <X11/Intrinsic.h>
 #include <X11/StringDefs.h>
 #include <X11/Shell.h>
-#include <X11/Xaw/Dialog.h>
-#include <X11/Xaw/Toggle.h>
+#include <X11/Xaw3d/Dialog.h>
+#include <X11/Xaw3d/Toggle.h>
 
+#include "global.h"
 #include "FileList.h"
-
 #include "Fm.h"
-
 #ifdef ENHANCE_HISTORY
 #include "Am.h" 	/* need to know aw.shell */
 #include "FmHistory.h"
@@ -42,18 +60,8 @@
 ---------------------------------------------------------------------------*/
 
 typedef struct {
-  FileWindowRec *fw;
-  Widget mkdir, createFile, goTo, rename, move, copy, link, select, filter;
-  char s[MAXPATHLEN], t[MAXPATHLEN];
-  char select_s[MAXPATHLEN];
-  char filter_s[MAXPATHLEN];
-  char mkdir_s[MAXPATHLEN];
-  char createFile_s[MAXPATHLEN];
-  char goTo_s[MAXPATHLEN];
-  char rename_s[MAXPATHLEN];
-  char move_s[MAXPATHLEN];
-  char copy_s[MAXPATHLEN];
-  char link_s[MAXPATHLEN];
+  struct PopupDialog *mkdir, *createFile, *goTo, *rename, *move, *copy, *link,
+		     *select, *filter;
 } PopupsRec;
 
 static PopupsRec popups;
@@ -62,7 +70,7 @@ static PopupsRec popups;
   PRIVATE FUNCTIONS
 ---------------------------------------------------------------------------*/
 
-char *dir_prefix(char *dir, char *path)
+static char *dir_prefix(char *dir, char *path)
 {
   char *p;
   if ((p = strrchr(path, '/'))) {
@@ -76,21 +84,25 @@ char *dir_prefix(char *dir, char *path)
   return dir;
 }
 
-static FmCallbackProc 
+static PDCallbackProc 
   mkdirOkCb, mkdirCancelCb, createFileOkCb, createFileCancelCb, goToOkCb,
   goToCancelCb, moveOkCb, moveCancelCb, copyOkCb, copyCancelCb, linkOkCb,
   linkCancelCb, selectAddCb, selectRemoveCb, selectCancelCb, selectReplaceCb,
   filterOkCb, filterClearCb, filterCancelCb;
 
-static void mkdirOkCb(Widget w, FileWindowRec *fw, XtPointer call_data)
+static void mkdirOkCb(UNUSED(Widget w), struct PopupDialog *p,
+                      UNUSED(void *call_data))
 {
-  XtPopdown(popups.mkdir);
-  fnexpand(popups.mkdir_s);
-  if (chdir(popups.fw->directory))
+  char *dir = getPopupAnswer(p,0);
+  FileWindowRec *fw = p->data;
+
+  popdownPopup(p);
+  fnexpand(dir);
+  if (chdir(fw->directory))
     sysError("System error:");
-  else if (mkdir(popups.mkdir_s, user.umask & 0777)) {
+  else if (mkdir(dir, user.umask & 0777)) {
     char s[0xff];
-    sprintf(s, "Error creating folder %s:", popups.mkdir_s);
+    sprintf(s, "Error creating folder %s:", dir);
     sysError(s);
   } else
     intUpdate();
@@ -99,23 +111,28 @@ static void mkdirOkCb(Widget w, FileWindowRec *fw, XtPointer call_data)
 
 /*---------------------------------------------------------------------------*/
 
-static void mkdirCancelCb(Widget w, FileWindowRec *fw, XtPointer call_data)
+static void mkdirCancelCb(UNUSED(Widget w), struct PopupDialog *p,
+                          UNUSED(void *call_data))
 {
-  XtPopdown(popups.mkdir);
+  popdownPopup(p);
   freeze = False;
 }
 
 /*---------------------------------------------------------------------------*/
 
-static void createFileOkCb(Widget w, FileWindowRec *fw, XtPointer call_data)
+static void createFileOkCb(UNUSED(Widget w), struct PopupDialog *p,
+		           UNUSED(void *call_data))
 {
-  XtPopdown(popups.createFile);
-  fnexpand(popups.createFile_s);
-  if (chdir(popups.fw->directory))
+  char *fn = getPopupAnswer(p,0);
+  FileWindowRec *fw = p->data;
+
+  popdownPopup(p);
+  fnexpand(fn);
+  if (chdir(fw->directory))
     sysError("System error:");
-  else if (create(popups.createFile_s, user.umask & 0666)) {
+  else if (create(fn, user.umask & 0666)) {
     char s[0xff];
-    sprintf(s, "Error creating file %s:", popups.createFile_s);
+    sprintf(s, "Error creating file %s:", fn);
     sysError(s);
   } else
     intUpdate();
@@ -124,55 +141,59 @@ static void createFileOkCb(Widget w, FileWindowRec *fw, XtPointer call_data)
 
 /*---------------------------------------------------------------------------*/
 
-static void createFileCancelCb(Widget w, FileWindowRec *fw, 
-			   XtPointer call_data)
+static void createFileCancelCb(UNUSED(Widget w), struct PopupDialog *p,
+                               UNUSED(void *call_data))
 {
-  XtPopdown(popups.createFile);
+  popdownPopup(p);
   freeze = False;
 }
 
 /*---------------------------------------------------------------------------*/
 
-static void goToOkCb(Widget w, FileWindowRec *fw, XtPointer call_data)
+static void goToOkCb(UNUSED(Widget w), struct PopupDialog *p,
+		     UNUSED(void *call_data))
 {
+  char *dir = getPopupAnswer(p,0);
+  FileWindowRec *fw = p->data;
   char path[MAXPATHLEN];
 #ifdef ENHANCE_SCROLL
   Boolean keep_position=True;
 #endif
 
-  XtPopdown(popups.goTo);
-  fnexpand(popups.goTo_s);
+  popdownPopup(p);
+  fnexpand(dir);
 #ifdef ENHANCE_BUGFIX
   /* if it's impossible to chdir(popups.fw->directory)
    * it still might be possible to chdir(popups.goTo_s) !
    * eg. if the actual directory was deleted, an we want to
    * change to another one.
    */
-  if (chdir(popups.goTo_s)) {
+  // TODO: this breaks chdir to relative directories, doesn't it?
+  if (chdir(dir)) {
     char s[0xff];
-    sprintf(s, "Can't open folder %s:", popups.goTo_s);
+    sprintf(s, "Can't open folder %s:", dir);
     sysError(s);
-	if (chdir(popups.fw->directory)) {
-    	  sprintf(s, "Can't open folder %s:", popups.fw->directory);
+	if (chdir(fw->directory)) {
+    	  sprintf(s, "Can't open folder %s:", fw->directory);
     	  sysError(s);
 	}
 #else
-  if (chdir(popups.fw->directory) || chdir(popups.goTo_s)) {
+  if (chdir(popups.fw->directory) || chdir(dir)) {
     char s[0xff];
-    sprintf(s, "Can't open folder %s:", popups.goTo_s);
+    sprintf(s, "Can't open folder %s:", dir);
     sysError(s);
 #endif
   } else if (!getwd(path))
     sysError("System error:");
   else {
 #ifdef ENHANCE_SCROLL
-    if (!(keep_position= (!strcmp(popups.fw->directory,path))))
+    if (!(keep_position= (!strcmp(fw->directory,path))))
 #endif
-    strcpy(popups.fw->directory, path);
+    strcpy(fw->directory, path);
 #ifdef ENHANCE_SCROLL
-    updateFileDisplay(popups.fw,keep_position);
+    updateFileDisplay(fw,keep_position);
 #else
-    updateFileDisplay(popups.fw);
+    updateFileDisplay(fw);
 #endif
   }
   freeze = False;
@@ -180,25 +201,28 @@ static void goToOkCb(Widget w, FileWindowRec *fw, XtPointer call_data)
 
 /*---------------------------------------------------------------------------*/
 
-static void goToCancelCb(Widget w, FileWindowRec *fw, XtPointer call_data)
+static void goToCancelCb(UNUSED(Widget w), struct PopupDialog *p,
+		         UNUSED(void *call_data))
 {
-  XtPopdown(popups.goTo);
+  popdownPopup(p);
   freeze = False;
 }
 
 /*---------------------------------------------------------------------------*/
 
-static void renameOkCb(Widget w, FileWindowRec *fw, XtPointer call_data)
+static void renameOkCb(UNUSED(Widget w), struct PopupDialog *p,
+		       UNUSED(void *call_data))
 {
+  FileWindowRec *fw = p->data;
   struct stat stats;
   int i;
   char *from = NULL, to[MAXPATHLEN], todir[MAXPATHLEN];
 
-  XtPopdown(popups.rename);
-  strcpy(to, popups.rename_s);
+  popdownPopup(p);
+  strcpy(to, getPopupAnswer(p,0));
   fnexpand(to);
 
-  if (chdir(popups.fw->directory))
+  if (chdir(fw->directory))
 
     sysError("System error:");
 
@@ -206,9 +230,9 @@ static void renameOkCb(Widget w, FileWindowRec *fw, XtPointer call_data)
 
     struct stat stats1;
 
-    for (i = 0; i < popups.fw->n_files; i++)
-      if (popups.fw->files[i]->selected) {
-	from = popups.fw->files[i]->name;
+    for (i = 0; i < fw->n_files; i++)
+      if (fw->files[i]->selected) {
+	from = fw->files[i]->name;
 	break;
       }
 
@@ -236,7 +260,7 @@ static void renameOkCb(Widget w, FileWindowRec *fw, XtPointer call_data)
       dir_prefix(todir, to);
       if ((*todir?chdir(todir):0) || !getwd(todir))
 	sysError("System error:");
-      markForUpdate(popups.fw->directory); markForUpdate(todir);
+      markForUpdate(fw->directory); markForUpdate(todir);
       intUpdate();
     }
 
@@ -248,26 +272,29 @@ static void renameOkCb(Widget w, FileWindowRec *fw, XtPointer call_data)
 
 /*---------------------------------------------------------------------------*/
 
-static void renameCancelCb(Widget w, FileWindowRec *fw, 
-			   XtPointer call_data)
+static void renameCancelCb(UNUSED(Widget w), struct PopupDialog *p,
+			   UNUSED(void *call_data))
 {
-  XtPopdown(popups.rename);
+  popdownPopup(p);
   freeze = False;
 }
 
 /*---------------------------------------------------------------------------*/
 
-static void moveOkCb(Widget w, FileWindowRec *fw, XtPointer call_data)
+static void moveOkCb(UNUSED(Widget w), struct PopupDialog *p,
+		     UNUSED(void *call_data))
 {
+  char *move = getPopupAnswer(p,0);
+  FileWindowRec *fw = p->data;
   struct stat stats;
   int i, toi, n_moved = 0;
   char *from = NULL, to[MAXPATHLEN], todir[MAXPATHLEN];
 
-  XtPopdown(popups.move);
-  strcpy(to, popups.move_s);
+  popdownPopup(p);
+  strcpy(to, move);
   fnexpand(to);
 
-  if (chdir(popups.fw->directory))
+  if (chdir(fw->directory))
 
     sysError("System error:");
 
@@ -278,10 +305,10 @@ static void moveOkCb(Widget w, FileWindowRec *fw, XtPointer call_data)
 
     if (!stat(to, &stats) && S_ISDIR(stats.st_mode)) {
 
-      if (chdir(to) || !getwd(to) || chdir(popups.fw->directory)) {
+      if (chdir(to) || !getwd(to) || chdir(fw->directory)) {
 	sysError("System error:");
 	goto out;
-      } else if (!strcmp(popups.fw->directory, to)) {
+      } else if (!strcmp(fw->directory, to)) {
 	error("Move:", "Source and destination are identical");
 	goto out;
       }
@@ -294,23 +321,24 @@ static void moveOkCb(Widget w, FileWindowRec *fw, XtPointer call_data)
 	to[toi] = '\0';
       }
 
-      for (i=0; i < popups.fw->n_files; i++)
-	if (popups.fw->files[i]->selected) {
-	  if (!strcmp(popups.fw->files[i]->name, ".") ||
-	      !strcmp(popups.fw->files[i]->name, "..")) {
+      for (i=0; i < fw->n_files; i++)
+	if (fw->files[i]->selected) {
+	  if (!strcmp(fw->files[i]->name, ".") ||
+	      !strcmp(fw->files[i]->name, "..")) {
 	    error("Cannot move . or ..", "");
 	    continue;
 	  }
-	  from = popups.fw->files[i]->name;
+	  from = fw->files[i]->name;
 	  strcpy(to+toi, from);
 	  if (exists(to) && resources.confirm_overwrite) {
 	    char s[0xff];
 	    sprintf(s, "Move: file %s already exists at destination", from);
-	    if (!confirm(s, "Overwrite?", ""))
+	    if (!confirm(s, "Overwrite?", "")) {
 	      if (aborted)
 		break;
 	      else
 		continue;
+	    }
 	  }
 	  if (rmove(from,to)) {
 	    char s[0xff];
@@ -324,7 +352,7 @@ static void moveOkCb(Widget w, FileWindowRec *fw, XtPointer call_data)
     /* otherwise only a single file may be selected; move it to the target
        file */
 
-    else if (popups.fw->n_selections > 1) {
+    else if (fw->n_selections > 1) {
 
       error("Move: target for multiple files", "must be a folder");
       goto out;
@@ -333,9 +361,9 @@ static void moveOkCb(Widget w, FileWindowRec *fw, XtPointer call_data)
 
       struct stat stats1;
 
-      for (i = 0; i < popups.fw->n_files; i++)
-	if (popups.fw->files[i]->selected) {
-	  from = popups.fw->files[i]->name;
+      for (i = 0; i < fw->n_files; i++)
+	if (fw->files[i]->selected) {
+	  from = fw->files[i]->name;
 	  break;
 	}
 
@@ -368,7 +396,7 @@ static void moveOkCb(Widget w, FileWindowRec *fw, XtPointer call_data)
     }
 
     if (n_moved) {
-      markForUpdate(popups.fw->directory); markForUpdate(todir);
+      markForUpdate(fw->directory); markForUpdate(todir);
       intUpdate();
     }
 
@@ -380,26 +408,29 @@ static void moveOkCb(Widget w, FileWindowRec *fw, XtPointer call_data)
 
 /*---------------------------------------------------------------------------*/
 
-static void moveCancelCb(Widget w, FileWindowRec *fw, 
-			   XtPointer call_data)
+static void moveCancelCb(UNUSED(Widget w), struct PopupDialog *p,
+			 UNUSED(void *call_data))
 {
-  XtPopdown(popups.move);
+  popdownPopup(p);
   freeze = False;
 }
 
 /*---------------------------------------------------------------------------*/
 
-static void copyOkCb(Widget w, FileWindowRec *fw, XtPointer call_data)
+static void copyOkCb(UNUSED(Widget w), struct PopupDialog *p,
+		     UNUSED(void *call_data))
 {
+  char *copy = getPopupAnswer(p,0);
+  FileWindowRec *fw = p->data;
   struct stat stats;
   int i, toi, n_copied = 0;
   char *from = NULL, to[MAXPATHLEN], todir[MAXPATHLEN];
 
-  XtPopdown(popups.copy);
-  strcpy(to, popups.copy_s);
+  popdownPopup(p);
+  strcpy(to, copy);
   fnexpand(to);
 
-  if (chdir(popups.fw->directory))
+  if (chdir(fw->directory))
 
     sysError("System error:");
 
@@ -410,10 +441,10 @@ static void copyOkCb(Widget w, FileWindowRec *fw, XtPointer call_data)
 
     if (!stat(to, &stats) && S_ISDIR(stats.st_mode)) {
 
-      if (chdir(to) || !getwd(to) || chdir(popups.fw->directory)) {
+      if (chdir(to) || !getwd(to) || chdir(fw->directory)) {
 	sysError("System error:");
 	goto out;
-      } else if (!strcmp(popups.fw->directory, to)) {
+      } else if (!strcmp(fw->directory, to)) {
 	error("Copy:", "Source and destination are identical");
 	goto out;
       }
@@ -426,23 +457,24 @@ static void copyOkCb(Widget w, FileWindowRec *fw, XtPointer call_data)
 	to[toi] = '\0';
       }
 
-      for (i=0; i < popups.fw->n_files; i++)
-	if (popups.fw->files[i]->selected) {
-	  if (!strcmp(popups.fw->files[i]->name, ".") ||
-	      !strcmp(popups.fw->files[i]->name, "..")) {
+      for (i=0; i < fw->n_files; i++)
+	if (fw->files[i]->selected) {
+	  if (!strcmp(fw->files[i]->name, ".") ||
+	      !strcmp(fw->files[i]->name, "..")) {
 	    error("Cannot copy . or ..", "");
 	    continue;
 	  }
-	  from = popups.fw->files[i]->name;
+	  from = fw->files[i]->name;
 	  strcpy(to+toi, from);
 	  if (exists(to) && resources.confirm_overwrite) {
 	    char s[0xff];
 	    sprintf(s, "Copy: file %s already exists at destination", from);
-	    if (!confirm(s, "Overwrite?", ""))
+	    if (!confirm(s, "Overwrite?", "")) {
 	      if (aborted)
 		break;
 	      else
 		continue;
+	    }
 	  }
 	  if (rcopy(from,to)) {
 	    char s[0xff];
@@ -456,7 +488,7 @@ static void copyOkCb(Widget w, FileWindowRec *fw, XtPointer call_data)
     /* otherwise only a single file may be selected; copy it to the target
        file */
 
-    else if (popups.fw->n_selections > 1) {
+    else if (fw->n_selections > 1) {
 
       error("Copy: target for multiple files", "must be a folder");
       goto out;
@@ -465,9 +497,9 @@ static void copyOkCb(Widget w, FileWindowRec *fw, XtPointer call_data)
 
       struct stat stats1;
 
-      for (i = 0; i < popups.fw->n_files; i++)
-	if (popups.fw->files[i]->selected) {
-	  from = popups.fw->files[i]->name;
+      for (i = 0; i < fw->n_files; i++)
+	if (fw->files[i]->selected) {
+	  from = fw->files[i]->name;
 	  break;
 	}
 
@@ -512,26 +544,28 @@ static void copyOkCb(Widget w, FileWindowRec *fw, XtPointer call_data)
 
 /*---------------------------------------------------------------------------*/
 
-static void copyCancelCb(Widget w, FileWindowRec *fw, 
-			   XtPointer call_data)
+static void copyCancelCb(UNUSED(Widget w), struct PopupDialog *p,
+			 UNUSED(void *call_data))
 {
-  XtPopdown(popups.copy);
+  popdownPopup(p);
   freeze = False;
 }
 
 /*---------------------------------------------------------------------------*/
 
-static void linkOkCb(Widget w, FileWindowRec *fw, XtPointer call_data)
+static void linkOkCb(UNUSED(Widget w), struct PopupDialog *p,
+		     UNUSED(void *call_data))
 {
+  FileWindowRec *fw = p->data;
   struct stat stats;
   int i, namei, toi, n_linked = 0;
   char *from = NULL, name[MAXPATHLEN], to[MAXPATHLEN], todir[MAXPATHLEN];
 
-  XtPopdown(popups.link);
-  strcpy(to, popups.link_s);
+  popdownPopup(p);
+  strcpy(to, getPopupAnswer(p,0));
   fnexpand(to);
 
-  strcpy(name, popups.fw->directory);
+  strcpy(name, fw->directory);
 
   namei = strlen(name);
   if (name[namei-1] != '/') {
@@ -539,7 +573,7 @@ static void linkOkCb(Widget w, FileWindowRec *fw, XtPointer call_data)
     name[namei] = '\0';
   }
 
-  if (chdir(popups.fw->directory))
+  if (chdir(fw->directory))
 
     sysError("System error:");
 
@@ -550,10 +584,10 @@ static void linkOkCb(Widget w, FileWindowRec *fw, XtPointer call_data)
 
     if (!stat(to, &stats) && S_ISDIR(stats.st_mode)) {
 
-      if (chdir(to) || !getwd(to) || chdir(popups.fw->directory)) {
+      if (chdir(to) || !getwd(to) || chdir(fw->directory)) {
 	sysError("System error:");
 	goto out;
-      } else if (!strcmp(popups.fw->directory, to)) {
+      } else if (!strcmp(fw->directory, to)) {
 	error("Link:", "Source and destination are identical");
 	goto out;
       }
@@ -566,19 +600,20 @@ static void linkOkCb(Widget w, FileWindowRec *fw, XtPointer call_data)
 	to[toi] = '\0';
       }
 
-      for (i=0; i < popups.fw->n_files; i++)
-	if (popups.fw->files[i]->selected) {
-	  from = popups.fw->files[i]->name;
+      for (i=0; i < fw->n_files; i++)
+	if (fw->files[i]->selected) {
+	  from = fw->files[i]->name;
 	  strcpy(name+namei, from);
 	  strcpy(to+toi, from);
 	  if (exists(to) && resources.confirm_overwrite) {
 	    char s[0xff];
 	    sprintf(s, "Link: file %s already exists at destination", from);
-	    if (!confirm(s, "Overwrite?", ""))
+	    if (!confirm(s, "Overwrite?", "")) {
 	      if (aborted)
 		break;
 	      else
 		continue;
+	    }
 	  }
 	  if (symlink(name,to)) {
 	    char s[0xff];
@@ -592,7 +627,7 @@ static void linkOkCb(Widget w, FileWindowRec *fw, XtPointer call_data)
     /* otherwise only a single file may be selected; link it to the target
        file */
 
-    else if (popups.fw->n_selections > 1) {
+    else if (fw->n_selections > 1) {
 
       error("Link: target for multiple files", "must be a folder");
       goto out;
@@ -601,9 +636,9 @@ static void linkOkCb(Widget w, FileWindowRec *fw, XtPointer call_data)
 
       struct stat stats1;
 
-      for (i = 0; i < popups.fw->n_files; i++)
-	if (popups.fw->files[i]->selected) {
-	  from = popups.fw->files[i]->name;
+      for (i = 0; i < fw->n_files; i++)
+	if (fw->files[i]->selected) {
+	  from = fw->files[i]->name;
 	  break;
 	}
 
@@ -647,10 +682,10 @@ static void linkOkCb(Widget w, FileWindowRec *fw, XtPointer call_data)
 
 /*---------------------------------------------------------------------------*/
 
-static void linkCancelCb(Widget w, FileWindowRec *fw, 
-			   XtPointer call_data)
+static void linkCancelCb(UNUSED(Widget w), struct PopupDialog *p,
+			 UNUSED(void *call_data))
 {
-  XtPopdown(popups.link);
+  popdownPopup(p);
   freeze = False;
 }
 
@@ -662,140 +697,151 @@ static void linkCancelCb(Widget w, FileWindowRec *fw,
 #define fnmatchnodot(pattern,fn) (strcmp(fn,".")&&strcmp(fn,"..")? \
 				  fnmatch(pattern,fn):!strcmp(pattern,fn))
 
-static void selectReplaceCb(Widget w, FileWindowRec *fw, 
-			    XtPointer call_data)
+static void selectReplaceCb(UNUSED(Widget w), struct PopupDialog *p,
+			    UNUSED(void *call_data))
 {
+  FileWindowRec *fw = p->data;
+  char *select_s = getPopupAnswer(p,0);
   int i;
   Pixel pix;
   Boolean new_val;
-  int onewid=XtIsSubclass(popups.fw->icon_box,fileListWidgetClass);
+  int onewid=XtIsSubclass(fw->icon_box,fileListWidgetClass);
 
-  XtPopdown(popups.select);
-  popups.fw->n_selections = 0;
-  popups.fw->n_bytes_selected = 0;
-  for (i=0; i<popups.fw->n_files; i++) {
-    if (onewid || popups.fw->files[i]->icon.toggle) {
-      if (fnmatchnodot(popups.select_s, popups.fw->files[i]->name)) {
+  popdownPopup(p);
+  fw->n_selections = 0;
+  fw->n_bytes_selected = 0;
+  for (i=0; i<fw->n_files; i++) {
+    if (onewid || fw->files[i]->icon.toggle) {
+      if (fnmatchnodot(select_s, fw->files[i]->name)) {
 	new_val = True;
-	popups.fw->n_selections++;
-	popups.fw->n_bytes_selected += popups.fw->files[i]->stats.st_size;
+	fw->n_selections++;
+	fw->n_bytes_selected += fw->files[i]->stats.st_size;
       }
       else
 	new_val = False;
-      if (new_val!= popups.fw->files[i]->selected) {
-	popups.fw->files[i]->selected = new_val;
+      if (new_val!= fw->files[i]->selected) {
+	fw->files[i]->selected = new_val;
         if (!onewid) {
-          XtVaGetValues(popups.fw->files[i]->icon.toggle,
-	  	      popups.fw->files[i]->selected?XtNforeground:XtNbackground,
+          XtVaGetValues(fw->files[i]->icon.toggle,
+	  	      fw->files[i]->selected?XtNforeground:XtNbackground,
 		      &pix, NULL);
-          XtVaSetValues(popups.fw->files[i]->icon.toggle, XtNborder,
+          XtVaSetValues(fw->files[i]->icon.toggle, XtNborder,
 		      (XtArgVal) pix, NULL);
         } else 
-	  FileListRefreshItem((FileListWidget)popups.fw->icon_box,i);
+	  FileListRefreshItem((FileListWidget)fw->icon_box,i);
      }
     }
   }
-  updateStatus(popups.fw);
+  updateStatus(fw);
   freeze = False;
 #ifdef ENHANCE_SELECTION
-  FmOwnSelection(popups.fw,CurrentTime);
+  FmOwnSelection(fw,CurrentTime);
 #endif
 }
 
 /*---------------------------------------------------------------------------*/
 
-static void selectAddCb(Widget w, FileWindowRec *fw, XtPointer call_data)
+static void selectAddCb(UNUSED(Widget w), struct PopupDialog *p,
+		        UNUSED(void *call_data))
 {
+  char *select_s = getPopupAnswer(p,0);
+  FileWindowRec *fw = p->data;
   int i;
   Pixel pix;
-  int onewid=XtIsSubclass(popups.fw->icon_box,fileListWidgetClass);
+  int onewid=XtIsSubclass(fw->icon_box,fileListWidgetClass);
   
-  XtPopdown(popups.select);
-  for(i=0; i<popups.fw->n_files; i++)
-    if (onewid || popups.fw->files[i]->icon.toggle) {
-      if (!popups.fw->files[i]->selected && 
-	  (fnmatchnodot(popups.select_s, popups.fw->files[i]->name))) {
-	popups.fw->files[i]->selected = True;
-	popups.fw->n_selections++;
-	popups.fw->n_bytes_selected += popups.fw->files[i]->stats.st_size;
+  popdownPopup(p);
+  for(i=0; i<fw->n_files; i++)
+    if (onewid || fw->files[i]->icon.toggle) {
+      if (!fw->files[i]->selected && 
+	  (fnmatchnodot(select_s, fw->files[i]->name))) {
+	fw->files[i]->selected = True;
+	fw->n_selections++;
+	fw->n_bytes_selected += fw->files[i]->stats.st_size;
 	if (!onewid) {
-	  XtVaGetValues(popups.fw->files[i]->icon.toggle, XtNforeground, &pix,
+	  XtVaGetValues(fw->files[i]->icon.toggle, XtNforeground, &pix,
 		      NULL);
-	  XtVaSetValues(popups.fw->files[i]->icon.toggle, XtNborder,
+	  XtVaSetValues(fw->files[i]->icon.toggle, XtNborder,
 		      (XtArgVal) pix, NULL);
-	} else FileListRefreshItem((FileListWidget)popups.fw->icon_box,i);
+	} else FileListRefreshItem((FileListWidget)fw->icon_box,i);
       }
     }
-  updateStatus(popups.fw);
+  updateStatus(fw);
   freeze = False;
 #ifdef ENHANCE_SELECTION
-  FmOwnSelection(popups.fw,CurrentTime);
+  FmOwnSelection(fw,CurrentTime);
 #endif
 }
 
 /*---------------------------------------------------------------------------*/
 
-static void selectRemoveCb(Widget w, FileWindowRec *fw, XtPointer call_data)
+static void selectRemoveCb(UNUSED(Widget w), struct PopupDialog *p,
+		           UNUSED(void *call_data))
 {
+  char *select_s = getPopupAnswer(p,0);
+  FileWindowRec *fw = p->data;
   int i;
   Pixel pix;
-  int onewid=XtIsSubclass(popups.fw->icon_box,fileListWidgetClass);
+  int onewid=XtIsSubclass(fw->icon_box,fileListWidgetClass);
   
-  XtPopdown(popups.select);
-  for(i=0; i<popups.fw->n_files; i++)
-    if (onewid || popups.fw->files[i]->icon.toggle) {
-      if (popups.fw->files[i]->selected && 
-	  (fnmatch(popups.select_s, popups.fw->files[i]->name))) {
-	popups.fw->files[i]->selected = False;
-	popups.fw->n_selections--;
-	popups.fw->n_bytes_selected -= popups.fw->files[i]->stats.st_size;
+  popdownPopup(p);
+  for(i=0; i<fw->n_files; i++)
+    if (onewid || fw->files[i]->icon.toggle) {
+      if (fw->files[i]->selected && 
+	  (fnmatch(select_s, fw->files[i]->name))) {
+	fw->files[i]->selected = False;
+	fw->n_selections--;
+	fw->n_bytes_selected -= fw->files[i]->stats.st_size;
 	if (!onewid) {
-	  XtVaGetValues(popups.fw->files[i]->icon.toggle, XtNbackground, &pix,
+	  XtVaGetValues(fw->files[i]->icon.toggle, XtNbackground, &pix,
 		      NULL);
-	  XtVaSetValues(popups.fw->files[i]->icon.toggle, XtNborder,
+	  XtVaSetValues(fw->files[i]->icon.toggle, XtNborder,
 		      (XtArgVal) pix, NULL);
-	} else FileListRefreshItem((FileListWidget)popups.fw->icon_box,i);
+	} else FileListRefreshItem((FileListWidget)fw->icon_box,i);
       }
     }
-  updateStatus(popups.fw);
+  updateStatus(fw);
   freeze = False;
 #ifdef ENHANCE_SELECTION
-  FmOwnSelection(popups.fw,CurrentTime);
+  FmOwnSelection(fw,CurrentTime);
 #endif
 }
 
 /*---------------------------------------------------------------------------*/
 
-static void selectCancelCb(Widget w, FileWindowRec *fw, 
-			   XtPointer call_data)
+static void selectCancelCb(UNUSED(Widget w), struct PopupDialog *p,
+			   UNUSED(void *call_data))
 {
-  XtPopdown(popups.select);
+  popdownPopup(p);
   freeze = False;
 }
 
 /*---------------------------------------------------------------------------*/
 
 /* KMR */
-static void filterCancelCb(Widget w, FileWindowRec *fw, 
-			   XtPointer call_data)
+static void filterCancelCb(UNUSED(Widget w), struct PopupDialog *p,
+			   UNUSED(void *call_data))
 {
-  XtPopdown(popups.filter);
+  popdownPopup(p);
   freeze = False;
 }
 
 /*---------------------------------------------------------------------------*/
 
 /* KMR */
-static void filterOkCb(Widget w, FileWindowRec *fw, 
-			   XtPointer call_data)
+static void filterOkCb(UNUSED(Widget w), struct PopupDialog *p,
+		       UNUSED(void *call_data))
 {
-  XtPopdown(popups.filter);
-  popups.fw->do_filter = True;
-  strcpy(popups.fw->dirFilter,popups.filter_s);
+  char *filter_s = getPopupAnswer(p,0);
+  FileWindowRec *fw = p->data;
+
+  popdownPopup(p);
+  fw->do_filter = True;
+  strcpy(fw->dirFilter,filter_s);
 #ifdef ENHANCE_SCROLL
-  updateFileDisplay(popups.fw,True);
+  updateFileDisplay(fw,True);
 #else
-  updateFileDisplay(popups.fw);
+  updateFileDisplay(fw);
 #endif
   freeze = False;
 }
@@ -803,16 +849,18 @@ static void filterOkCb(Widget w, FileWindowRec *fw,
 /*---------------------------------------------------------------------------*/
 
 /* KMR */
-static void filterClearCb(Widget w, FileWindowRec *fw, 
-			   XtPointer call_data)
+static void filterClearCb(UNUSED(Widget w), struct PopupDialog *p,
+			  UNUSED(void *call_data))
 {
-  XtPopdown(popups.filter);
-  popups.fw->do_filter = False;
-  popups.fw->dirFilter[0] = '\0';
+  FileWindowRec *fw = p->data;
+
+  popdownPopup(p);
+  fw->do_filter = False;
+  fw->dirFilter[0] = '\0';
 #ifdef ENHANCE_SCROLL
-  updateFileDisplay(popups.fw,True);
+  updateFileDisplay(fw,True);
 #else
-  updateFileDisplay(popups.fw);
+  updateFileDisplay(fw);
 #endif
   freeze = False;
 }
@@ -821,49 +869,49 @@ static void filterClearCb(Widget w, FileWindowRec *fw,
   Button information for popups
 ---------------------------------------------------------------------------*/
 
-static ButtonRec mkdir_buttons[] = {
+static const PDButtonRec mkdir_buttons[] = {
   { "ok", "Ok", mkdirOkCb },
   { "cancel", "Cancel", mkdirCancelCb }
 };
 
-static ButtonRec createFile_buttons[] = {
+static const PDButtonRec createFile_buttons[] = {
   { "ok", "Ok", createFileOkCb },
   { "cancel", "Cancel", createFileCancelCb }
 };
 
-static ButtonRec goTo_buttons[] = {
+static const PDButtonRec goTo_buttons[] = {
   { "ok", "Ok", goToOkCb },
   { "cancel", "Cancel", goToCancelCb }
 };
 
-static ButtonRec rename_buttons[] = {
+static const PDButtonRec rename_buttons[] = {
   { "ok", "Ok", renameOkCb },
   { "cancel", "Cancel", renameCancelCb }
 };
 
-static ButtonRec move_buttons[] = {
+static const PDButtonRec move_buttons[] = {
   { "ok", "Ok", moveOkCb },
   { "cancel", "Cancel", moveCancelCb }
 };
 
-static ButtonRec copy_buttons[] = {
+static const PDButtonRec copy_buttons[] = {
   { "ok", "Ok", copyOkCb },
   { "cancel", "Cancel", copyCancelCb }
 };
 
-static ButtonRec link_buttons[] = {
+static const PDButtonRec link_buttons[] = {
   { "ok", "Ok", linkOkCb },
   { "cancel", "Cancel", linkCancelCb }
 };
 
-static ButtonRec select_buttons[] = {
+static const PDButtonRec select_buttons[] = {
   { "replace", "Replace", selectReplaceCb },
   { "add", "Add", selectAddCb },
   { "remove", "Remove", selectRemoveCb },
   { "cancel", "Cancel", selectCancelCb }
 };
 
-static ButtonRec filter_buttons[] = {       /* KMR */
+static const PDButtonRec filter_buttons[] = {       /* KMR */
   { "ok", "Ok", filterOkCb },
   { "clear", "Clear", filterClearCb },
   { "cancel", "Cancel", filterCancelCb }
@@ -873,40 +921,40 @@ static ButtonRec filter_buttons[] = {       /* KMR */
   Question information for popups
 ---------------------------------------------------------------------------*/
 
-static QuestionRec mkdir_questions[] = {
-  { "Create folder:", popups.mkdir_s, MAXPATHLEN, NULL }
+static const QuestionRec mkdir_questions[] = {
+  { "Create folder:", MAXPATHLEN }
 };
 
-static QuestionRec createFile_questions[] = {
-  { "Create file:", popups.createFile_s, MAXPATHLEN, NULL }
+static const QuestionRec createFile_questions[] = {
+  { "Create file:", MAXPATHLEN }
 };
 
-static QuestionRec goTo_questions[] = {
-  { "Go to folder:", popups.goTo_s, MAXPATHLEN, NULL }
+static const QuestionRec goTo_questions[] = {
+  { "Go to folder:", MAXPATHLEN }
 };
 
-static QuestionRec rename_questions[] = {
-  { "Rename to:", popups.rename_s, MAXPATHLEN, NULL }
+static const QuestionRec rename_questions[] = {
+  { "Rename to:", MAXPATHLEN }
 };
 
-static QuestionRec move_questions[] = {
-  { "Move to:", popups.move_s, MAXPATHLEN, NULL }
+static const QuestionRec move_questions[] = {
+  { "Move to:", MAXPATHLEN }
 };
 
-static QuestionRec copy_questions[] = {
-  { "Copy to:", popups.copy_s, MAXPATHLEN, NULL }
+static const QuestionRec copy_questions[] = {
+  { "Copy to:", MAXPATHLEN }
 };
 
-static QuestionRec link_questions[] = {
-  { "Link to:", popups.link_s, MAXPATHLEN, NULL }
+static const QuestionRec link_questions[] = {
+  { "Link to:", MAXPATHLEN }
 };
 
-static QuestionRec select_questions[] = {
-  { "Filename pattern:", popups.select_s, MAXPATHLEN, NULL }
+static const QuestionRec select_questions[] = {
+  { "Filename pattern:", MAXPATHLEN }
 };
 
-static QuestionRec filter_questions[] = {             /* KMR */
-  { "Filename pattern:", popups.filter_s, MAXPATHLEN, NULL }
+static const QuestionRec filter_questions[] = {             /* KMR */
+  { "Filename pattern:", MAXPATHLEN }
 };
 
 #ifdef ENHANCE_HISTORY
@@ -1022,124 +1070,104 @@ void createMainPopups()
 
 /*---------------------------------------------------------------------------*/ 
 
-void selectPopup(Widget w, FileWindowRec *fw, XtPointer call_data)
+void selectPopup(UNUSED(Widget w), FileWindowRec *fw, UNUSED(void *call_data))
 {
-  popups.fw = fw;
-  XtVaSetValues(select_questions[0].widget, XtNstring, 
-		(XtArgVal) popups.select_s, NULL);
+  reloadPopupData(popups.select);
   freeze = True;
-  popupByCursor(popups.select, XtGrabExclusive);
+  popupPopupByCursor(popups.select, fw);
 }
 
 /*---------------------------------------------------------------------------*/ 
 
 /* KMR */
-void filterPopup(Widget w, FileWindowRec *fw, XtPointer call_data)
+void filterPopup(UNUSED(Widget w), FileWindowRec *fw, UNUSED(void *call_data))
 {
-  popups.fw = fw;
-  strcpy(popups.filter_s,fw->dirFilter);
-  XtVaSetValues(filter_questions[0].widget, XtNstring, 
-		(XtArgVal) popups.filter_s, NULL);
+  setPopupData(popups.filter,0,fw->dirFilter);
   freeze = True;
-  popupByCursor(popups.filter, XtGrabExclusive);
+  popupPopupByCursor(popups.filter, fw);
 }
 
 /*---------------------------------------------------------------------------*/
 
-void mkdirPopup(Widget w, FileWindowRec *fw, XtPointer call_data)
+void mkdirPopup(UNUSED(Widget w), FileWindowRec *fw, UNUSED(void *call_data))
 {
-  popups.fw = fw;
-  XtVaSetValues(mkdir_questions[0].widget, XtNstring, 
-		(XtArgVal) popups.mkdir_s, NULL);
+  reloadPopupData(popups.mkdir);
   freeze = True;
-  popupByCursor(popups.mkdir, XtGrabExclusive);
+  popupPopupByCursor(popups.mkdir, fw);
 }
 
 /*---------------------------------------------------------------------------*/
 
-void createFilePopup(Widget w, FileWindowRec *fw, XtPointer call_data)
+void createFilePopup(UNUSED(Widget w), FileWindowRec *fw, UNUSED(void *call_data))
 {
-  popups.fw = fw;
-  XtVaSetValues(createFile_questions[0].widget, XtNstring, 
-		(XtArgVal) popups.createFile_s, NULL);
+  reloadPopupData(popups.createFile);
   freeze = True;
-  popupByCursor(popups.createFile, XtGrabExclusive);
+  popupPopupByCursor(popups.createFile, fw);
 }
 
 /*---------------------------------------------------------------------------*/
 
-void goToPopup(Widget w, FileWindowRec *fw, XtPointer call_data)
+void goToPopup(UNUSED(Widget w), FileWindowRec *fw, UNUSED(void *call_data))
 {
-  popups.fw = fw;
-  XtVaSetValues(goTo_questions[0].widget, XtNstring, 
-		(XtArgVal) popups.goTo_s, NULL);
+  reloadPopupData(popups.goTo);
   freeze = True;
-  popupByCursor(popups.goTo, XtGrabExclusive);
+  popupPopupByCursor(popups.goTo, fw);
 }
 
 /*---------------------------------------------------------------------------*/
 
-void renamePopup(Widget w, FileWindowRec *fw, XtPointer call_data)
+void renamePopup(UNUSED(Widget w), FileWindowRec *fw, UNUSED(void *call_data))
 {
   int i;
 
-  if (fw == NULL) fw = popup_fw;
+  assert( fw != NULL );
 
   if (!fw->n_selections) return;
 
-  popups.fw = fw;
-  for (i = 0; i < popups.fw->n_files; i++)
-    if (popups.fw->files[i]->selected) {
-      strcpy(popups.rename_s, popups.fw->files[i]->name);
+  for (i = 0; i < fw->n_files; i++)
+    if (fw->files[i]->selected) {
+      setPopupData(popups.rename,0,fw->files[i]->name);
       break;
     }
-  XtVaSetValues(rename_questions[0].widget, XtNstring, 
-		(XtArgVal) popups.rename_s, NULL);
   freeze = True;
-  popupByCursor(popups.rename, XtGrabExclusive);
+  popupPopupByCursor(popups.rename, fw);
 }
 
 /*---------------------------------------------------------------------------*/
 
-void movePopup(Widget w, FileWindowRec *fw, XtPointer call_data)
+void movePopup(UNUSED(Widget w), FileWindowRec *fw, UNUSED(void *call_data))
 {
-  if (fw == NULL) fw = popup_fw;
+  assert( fw != NULL );
 
   if (!fw->n_selections) return;
 
-  popups.fw = fw;
-  XtVaSetValues(move_questions[0].widget, XtNstring, 
-		(XtArgVal) popups.move_s, NULL);
+  reloadPopupData(popups.move);
   freeze = True;
-  popupByCursor(popups.move, XtGrabExclusive);
+  popupPopupByCursor(popups.move, fw);
 }
 
 /*---------------------------------------------------------------------------*/
 
-void copyPopup(Widget w, FileWindowRec *fw, XtPointer call_data)
+void copyPopup(UNUSED(Widget w), FileWindowRec *fw, UNUSED(void *call_data))
 {
-  if (fw == NULL) fw = popup_fw;
+  assert( fw != NULL );
 
   if (!fw->n_selections) return;
 
-  popups.fw = fw;
-  XtVaSetValues(copy_questions[0].widget, XtNstring, 
-		(XtArgVal) popups.copy_s, NULL);
+  reloadPopupData(popups.copy);
   freeze = True;
-  popupByCursor(popups.copy, XtGrabExclusive);
+  popupPopupByCursor(popups.copy, fw);
 }
 
 /*---------------------------------------------------------------------------*/
 
-void linkPopup(Widget w, FileWindowRec *fw, XtPointer call_data)
+void linkPopup(UNUSED(Widget w), FileWindowRec *fw, UNUSED(void *call_data))
 {
-  if (fw == NULL) fw = popup_fw;
+  assert( fw != NULL );
 
   if (!fw->n_selections) return;
 
-  popups.fw = fw;
-  XtVaSetValues(link_questions[0].widget, XtNstring, 
-		(XtArgVal) popups.link_s, NULL);
+  reloadPopupData(popups.link);
   freeze = True;
-  popupByCursor(popups.link, XtGrabExclusive);
+  popupPopupByCursor(popups.link, fw);
 }
